# -*- coding: utf-8 -*-
"""
Feature class for the time series of profiles observation pattern.
"""
import typing as T

import xarray as xr

import cerbere
from cerbere.feature.cbasecollection import BaseCollection
from cerbere.feature.cimdcollection import IMDCollection
from cerbere.feature.comdcollection import OMDCollection


class TimeSeriesProfile(BaseCollection):
    """
    This class is for a single time series. This feature is defined by
    geolocation coordinates field: ``lon()``, ``lat()``, ``z()``,
    ``time(time)``,  none having a geolocation dimension (except ``time`` for T
    axis coordinate).

    Refer to CF convention: https://cfconventions.org/cf-conventions/cf
    -conventions.html#time-series-ptofiles

    When profiles are taken repeatedly at a station, one gets a time series
    of profiles (see also section H.2 for discussion of stations and time
    series). The resulting collection of profiles is called a
    timeSeriesProfile. A data variable may contain a collection of such
    timeSeriesProfile features, one feature per station. The instance
    dimension in the case of a timeSeriesProfile is also referred to as the
    station dimension. The instance variables, which have just this
    dimension, including latitude and longitude for example, are also
    referred to as station variables and are considered to contain
    information describing the stations. The station variables may contain
    missing values. This allows one to reserve space for additional stations
    that may be added at a later time, as discussed in section 9.6. In addition,

    * It is strongly recommended that there should be a station variable (
      which may be of any type) with cf_role attribute "timeseries_id",
      whose values uniquely identify the stations.
    * It is recommended that there should be station variables with
      standard_name attributes "platform_name", "surface_altitude" and
     “platform_id” when applicable.

    TimeSeriesProfiles are more complicated than timeSeries because there are
    two element dimensions (profile and vertical). Each time series has a
    number of profiles from different times as its elements, and each profile
    has a number of data from various levels as its elements. It is strongly
    recommended that there always be a variable (of any data type) with the
    profile dimension and the cf_role attribute " profile_id ", whose values
    uniquely identify the profiles.

    A :class:`~cerbere.feature.timeseriesprofile.TimeSeriesProfile` feature is
    defined by geolocation coordinates field, , all having a single geolocation
    dimension: ``profile`` (except ``z`` for Z axis coordinate).
      * ``lon(profile)``
      * ``lat(profile)``
      * ``time(profile)``
      * ``depth/alt (profile, z)``
    """
    cf_instance_axis = 'station'
    cf_element_instance_axis = 'profile'
    cf_instance_role = 'timeseries_id'
    cf_feature_type = 'timeSeriesProfile'
    cdm_data_type = 'station_profile'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': (cf_instance_axis,),
        'latitude': (cf_instance_axis,),
        'time': (cf_instance_axis, 'profile'),
        'vertical': [(cf_instance_axis, 'profile', 'Z'),
                     (cf_instance_axis, 'Z')],
        'data': (cf_instance_axis, 'profile', 'Z')
    }
    # optional axes for the feature
    cf_optional_axes = []

    def __init__(self, *args, **kwargs):
        kwargs['instance_class'] = 'Profile'
        super().__init__(*args, **kwargs)

        # self.cf_role_var = self.cf_role()

        # CF specific attrs and vars for profile feature
        self.ds.attrs['featureType'] = 'timeSeriesProfile'

    @classmethod
    def is_alias(cls, dataset):
        # @TODO clarify if and which collection
        aliases = [
            _ for _ in [IMDCollection, OMDCollection]
            if _.guess_feature(dataset) is not None]
        if len(aliases) > 0:
            return aliases[0]

    # @classmethod
    # def _canonical_dims(cls, coord: str, instance_class=None) -> T.Set[str]:
    #     return set(cls.cf_canonical_dims[coord]) - {cls.cf_instance_axis}

    @classmethod
    def _canonical_dims(cls, coord: str, instance_class=None) -> T.Set[str]:
        if coord in ['latitude', 'longitude', 'time']:
            return set(cls.cf_canonical_dims[coord]) - {cls.cf_instance_axis}

        return instance_class.cf_canonical_dims[coord]

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[IMDCollection]:

        cfdst = dataset.cb.cfdataset

        # @TODO: check role attributes

        # check CF featureType and Unidata CDM' feature type are compatible
        # with this feature
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        try:
            cls._check_feature(dataset, instance_class='Profile')
            return cls
        except cerbere.AxisError:
            pass


class UniZTimeSeriesProfile(OMDCollection):
    """
    Feature class for the time series of profiles observation pattern, where z
    is unique and identical for each profile in the series.

    A :class:`~cerbere.feature.timeseriesprofile.TimeSeriesProfile` feature is
    defined by geolocation coordinates field, all having a single geolocation
    dimension: ``profile`` (except ``z`` for Z axis coordinate).
      * ``lon(profile)``
      * ``lat(profile)``
      * ``time(profile)``
      * ``depth/alt (z)``
    """
    cf_instance_axis = 'timeseries_id'
    cf_feature_type = 'timeSeriesProfile'
    cdm_data_type = 'timeseries_profile'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'X': (cf_instance_axis,),
        'Y': (cf_instance_axis,),
        'T': (cf_instance_axis, 'profile'),
        'Z': (cf_instance_axis, 'Z'),
        'data': (cf_instance_axis, 'profile', 'Z')
    }

    def __init__(self, *args, **kwargs):
        kwargs['instance_class'] = 'Profile'
        super().__init__(*args, **kwargs)

        #self.cf_role_var = self.ds.cf_role()

        # CF specific attrs and vars for profile feature
        self.ds.attrs['featureType'] = 'timeSeriesProfile'

    @classmethod
    def is_alias(cls):
        return OMDCollection

    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[IMDCollection]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if cls._unexpected_featureType(cfdst) or \
                cls._unexpected_cdm_data_type(cfdst):
            return

        # check lat/lon/time are axes with a single dimension which profile
        for coordname in ['lon', 'lat', 'time']:
            if len(cfdst[coordname].dims) != 0:
                return

        # check provided axes if any
        if cfdst.cb.Z is not None:
            if len(cfdst[cfdst.cb.Z.name].dims) != 2 and \
                    cfdst[cfdst.cb.Z.name].dims[0] != 'profile':
                return

        return cls


# -*- coding: utf-8 -*-
"""
Feature classes for grid observation patterns
"""
from typing import Tuple

import numpy as np
import xarray as xr

from .feature import Feature

__all__ = ['DEFAULT_PROJECTION', 'Projection', 'Grid', 'CylindricalGrid']


# Plate Carree projection as default (regular isolat, isolon)
DEFAULT_PROJECTION = (
    '+proj=eqc +lat_ts=0 +lat_0=0 +lon_0=0 +x_0=0 +y_0=0 +a=6378137 +b=6378137'
    ' +units=m'
)


class Projection(object):
    """Definition of a grid projection"""
    def __init__(self,
                 proj4_definition: str = DEFAULT_PROJECTION,
                 identifier: str = 'regular'
                 ):
        self.definition = proj4_definition
        self.identifier = identifier

    def is_cylindrical(self) -> bool:
        """True if the projection is cylindrical"""
        return self.identifier == 'regular'


class Grid(Feature):
    """
    Model class for the grid feature, ie a two-dimensional array on fixed
    projection, resolution and boundaries.

    A grid can be built from the ``content`` argument:
        * if None, the feature will undefined but can be updated later (for
          instance loading the content from a file with ``load`` method)
        * from the content of a file, providing a mapper class object
        * from an xarray dataset object
        * from a dictionary of coordinate and data fields. For coordinate
          fields, use only any of 'lat', 'lon', 'time', 'z' as keys.

          The coordinate fields can be provided as numpy arrays of values,
          :class:`cerbere.datamodel.field.Field` class objects, or
          :class:`xarray.DataArray` class objects.

          The lat and lon coordinate fields can also be provided as tuple
          (min, max, step): this is only for a regular grid. The values
          are automatically generated.

          The geophysical fields, in ``fields`` can be
          :class:`cerbere.datamodel.field.Field` class objects, or
          :class:`xarray.DataArray` class objects.

    .. code-block:: python

        from cerbere.feature.grid import Grid
        from cerbere.dataset.ncdataset import NCDataset

        ncf = NCDataset(url='SSS_SMOS_L3_Daily_0.5deg_CATDS_CECOS_2012.12.30_V02.nc')
        g = Grid(ncf)

    Args:
        content (xarray, mapper or dict of fields): the content from which to
            build the feature coordinates and data fields.

        attrs (dict): a dictionary of global attributes (as in a NetCDF file).
            Make use of the CF attributes. If the feature content is read from
            a file through a mapper, these attributes will be added to the
            existing the list of attributes already in the file (or replace the
            values of the existing attributes with the same name in the file).
            Use with caution in this context.
    """
    _required_dims = 'y', 'x',

    def __init__(self,
                 *args,
                 projection: 'Projection' = Projection(),
                 **kwargs
                 ):
        """
        """
        # projection parameters
        self.projection = projection

        # create feature
        super(Grid, self).__init__(
            *args,
            **kwargs
        )

        # squeeze time dimension in non-coordinate fields
        ctime = self._std_dataset.time.copy()
        if 'time' in self._std_dataset.dims:
            self._std_dataset = self._std_dataset.squeeze(dim='time')
        self._std_dataset.coords['time'] = ctime

    def get_geocoord_dimnames(
            self, fieldname: str,
            values: 'xr.DataArray') -> Tuple[str, ...]:
        if fieldname in ['depth', 'height', 'alt']:
            return 'z',
        elif fieldname == 'time':
            if len(values.shape) == 1:
                return 'time',
            else:
                return 'y', 'x'
        else:
            return 'y', 'x'

    def get_spatial_resolution(self):
        """Return the spatial resolution of the feature, in degrees"""
        if self.spatial_resolution is None:
            mapper = self.get_mapper()
            if mapper:
                return mapper.get_spatial_resolution_in_deg()
        else:
            return self.spatial_resolution



class CylindricalGrid(Grid):

    def __init__(self,
                 *args,
                 **kwargs
                 ):
        """
        Cylindrical lat/lon grid.
        """
        # create feature
        super(CylindricalGrid, self).__init__(
            *args,
            **kwargs
        )

    @property
    def _feature_geodimnames(self):
        return 'lat', 'lon',

    def get_geocoord_dimnames(self, fieldname, values):
        if fieldname in ['depth', 'height']:
            if len(values.shape) != 1:
                raise ValueError('z coordinate must be one-dimensional')
            return ('z',)
        elif fieldname == 'time':
            if len(values.shape) == 1:
                return ('time',)
            elif len(values.shape) == 2:
                return ('lat', 'lon',)
            else:
                raise ValueError(
                    'time coordinate must be one or two dimensional'
                )
        else:
            return ('lat', 'lon',)

    def closest_spatial_location(self, lon, lat):
        """Get closest dataset lat/lon location to given coordinates.

        Use pythagorian differences on lat/lon values so take the result with
        caution.

        Returns:
            A tuple with the indices and lon/lat of the closest point found.
        """
        idx_lon = np.abs(self.get_lon() - lon).argmin()
        idx_lat = np.abs(self.get_lat() - lat).argmin()

        loc = {'lat': idx_lat, 'lon': idx_lon}

        geoloc = (self.get_lon().flat[idx_lon], self.get_lat().flat[idx_lat])

        return loc, geoloc,



# -*- coding: utf-8 -*-
"""
Feature class for the section observation pattern (a time series of profiles).
"""
from __future__ import absolute_import, division, print_function

from typing import Tuple

from cerbere.feature.feature import Feature

__all__ = ['TrajectoryProfile']


class TrajectoryProfile(Feature):
    """
    Feature class for the section observation patterns.

    A :class:`~cerbere.feature.section.Section` feature is defined by
    geolocation coordinates field: ``lon(profile)``, ``lat(profile)``,
    ``time(profile)``, ``depth/alt (profile, z)``, all having a single
    geolocation dimension: ``profile`` (except ``z`` for Z axis coordinate).
    """
    _feature_geodimnames = 'profile', 'z',

    def __init__(self, *args, **kwargs):
        super(TrajectoryProfile, self).__init__(*args, **kwargs)

        # CF specific attrs and vars for profile feature
        self._std_dataset.attrs['featureType'] = 'TrajectoryProfile'

        # mandatory coordinates
        for _, v in self._std_dataset.data_vars.items():
            v.attrs['coordinates'] = 'time lat lon z'

    def get_geocoord_dimnames(
            self, fieldname: str, *args, **kwargs) -> Tuple[str, ...]:
        if fieldname in ['lat', 'lon', 'time']:
            return 'profile',
        else:
            return 'profile', 'z',



"""
Test methods for a xarray/cerbere reader classes for specific datasets.

All these methods must be imported in the test module for the new
reader. This test module must implement the following fixtures:

* test_file : return the input file to be used in this test
* download_url : return the url to test data repository
* reader : return the reader class name
* feature : return the related feature class corresponding to the dataset
"""
import datetime
import os
from pathlib import Path

import pytest

import numpy
import xarray as xr

import cerbere
from cerbere.feature.cbasefeature import BaseFeature
from cerbere.feature.cbasecollection import BaseCollection


TEST_SUBSET = 'test_subset.nc'


@pytest.fixture(scope='module')
def download_url():
    return 'ftp://ftp.ifremer.fr/ifremer/cersat/projects/cerbere/test_data/'


@pytest.fixture(scope='module')
def input_file(test_file, reader, download_url):
    # search test file
    if 'CERBERE_TESTDATA_REPOSITORY' not in os.environ:
        rootpath = Path('tests/data/')
    else:
        rootpath = Path(os.environ['CERBERE_TESTDATA_REPOSITORY'])
    dirname = Path(reader.split('.')[-1])
    testfilepath = rootpath / dirname / test_file

    if not testfilepath.exists():
        raise IOError(
            f'Test file {test_file} was not found. Please download it from '
              f'{download_url} and place in {Path(rootpath /dirname)} folder')

    return testfilepath


@pytest.fixture(scope='module')
def dataset(input_file, reader):
    """Open file and create dataset object"""
    return cerbere.open_dataset(input_file, reader=reader)


@pytest.fixture(scope='module')
def feature_class():
    return


@pytest.fixture(scope='module')
def feature(input_file, reader):
    return cerbere.open_feature(input_file, reader=reader)


@pytest.fixture(scope='module')
def output_dir(tmpdir_factory):
    return tmpdir_factory.mktemp('output')


def test_open_file_as_dataset(dataset, input_file):
    """Test dataset instantiation with an input file"""
    assert isinstance(dataset, xr.Dataset)
    assert dataset.cb.url == input_file


def test_open_file_as_feature(feature, feature_class):
    """Test dataset instantiation with an input file"""
    assert isinstance(feature, feature_class)

    if issubclass(feature_class, BaseCollection):
        assert feature.instance_class is not None


def test_open_dataset_and_feature(input_file, feature_class, reader):
    feature = cerbere.open_feature(
        input_file, feature_class.__name__, reader)
    assert isinstance(feature, BaseFeature)


def test_get_list_of_vars(dataset):
    """Test reading the list of fields"""
    for v in dataset.data_vars:
        assert isinstance(v, str)
        assert (v != "")


def test_read_vars(dataset):
    """Test reading and displaying variable properties"""
    vars = dataset.data_vars
    for v in vars:
        print(f'...reading {v}')
        assert isinstance(dataset.data_vars[v], xr.DataArray)
        print(dataset.data_vars[v])


def test_read_global_dimensions(dataset):
    """Test reading and displaying global dimensions"""
    print('...reading global dimensions')
    dims = dataset.dims

    assert dims is not None
    assert len(dims) > 0

    for dim in dims:
        dimsize = dataset.sizes[dim]
        print('.....', dim, dimsize)

        assert isinstance(dim, str)
        assert isinstance(dimsize, int)


def test_read_lat(dataset):
    """Test reading latitude and checking validity"""
    print('fillvalue', dataset['lat'].cb.fill_value)
    data = dataset['lat'].to_masked_array()

    print('...test there are no fillvalues')
    assert data.size == data.count()

    print('...test values are in valid range')
    assert -90 <= data.min() <= 90
    assert -90 <= data.max() <= 90


def test_read_lon(dataset):
    """Test reading longitude and checking validity"""
    data = dataset['lon'].to_masked_array()

    print('...test there are no fillvalues')
    assert data.size == data.count()

    print('...test values are in valid range')
    assert -180 <= data.min() <= 180
    assert -180 <= data.max() <= 180


def test_read_time(dataset):
    """Test reading time and checking validity"""
    def dt64todatetime(dt64):
        unix_epoch = numpy.datetime64(0, 's')
        one_second = numpy.timedelta64(1, 's')
        seconds_since_epoch = (dt64 - unix_epoch) / one_second
        return datetime.datetime.utcfromtimestamp(seconds_since_epoch)

    data = dataset['time'].to_masked_array()

    print('...test there are no fillvalues')
    assert data.size == data.count()
    assert isinstance(data.min(), numpy.datetime64)

    print('...test values are in valid range')
    time_min = dt64todatetime(data.min())
    print(data.min(), data.min().astype(datetime.datetime))
    print(f'...min time: {time_min} {type(time_min)}, {data.min()} '
          f'{type(data.min())}')
    assert time_min > datetime.datetime(1970, 1, 1)

    time_max = dt64todatetime(data.max())
    print('...max time: ', time_max)
    assert time_max < datetime.datetime(2050, 1, 1)


def test_read_global_attributes(dataset):
    """Test reading the global attributes"""
    for attr in dataset.attrs:
        print(attr, dataset.attrs[attr])
        assert isinstance(dataset.attrs[attr],
                          (str, int, datetime.datetime, numpy.int32, numpy.int8, numpy.int64,
                           numpy.uint32, numpy.int16, numpy.float32,
                           numpy.float64, list))


def test_read_time_coverage(dataset):
    """Test reading the coverage start end end time"""
    start = dataset.cb.time_coverage_start
    print('...start time : ', start)
    assert isinstance(start, datetime.datetime)

    end = dataset.cb.time_coverage_end
    print('...end time : ', end)
    assert isinstance(end, datetime.datetime)


def __extract_subset(feature, feature_class):
    """Extract a subset from a datamodel structure loaded with a file"""
    datasetobj = feature.dataset

    if feature.__class__.__name__ in ['Swath', 'Image']:
        cells = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['X']]
        rows = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['Y']]
        width = min(min(rows // 2, cells // 2), 5)
        r0, r1 = rows // 2 - width, rows // 2 + width
        c0, c1 = cells // 2 - width, cells // 2 + width
        slices = {feature.ds.cb.cf_axis_dims['Y']: slice(r0, r1, 1),
                  feature.ds.cb.cf_axis_dims['X']: slice(c0, c1, 1)}
        print('Subset: ', slices)
        subset = feature.isel(slices)

    elif feature.__class__.__name__ in ['Grid', 'GridTimeSeries']:
        if feature.is_cylindrical():
            ni = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['X']]
            nj = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['Y']]
            width = min(min(nj // 2, ni // 2), 5)
            j0, j1 = nj // 2 - width, nj // 2 + width
            i0, i1 = ni // 2 - width, ni // 2 + width
            subset = feature.isel(
                {'lat': slice(j0, j1, 1), 'lon': slice(i0, i1, 1)})
        else:
            ni = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['X']]
            nj = feature.ds.cb.cf_dims[feature.ds.cb.cf_axis_dims['Y']]
            width = min(min(nj // 2, ni // 2), 5)
            j0, j1 = nj // 2 - width, nj // 2 + width
            i0, i1 = ni // 2 - width, ni // 2 + width
            subset = feature.isel(
                {'y': slice(j0, j1, 1), 'x': slice(i0, i1, 1)})

    elif feature.__class__.__name__ in ['Trajectory']:
        times = feature.ds.dims[feature.ds.cb.cf_axis_dims['T']]
        width = min(times // 2, 5)
        r0, r1 = times // 2 - width, times // 2 + width
        print('Subset ')
        print('time : ', r0, r1)
        subset = feature.isel({'time': slice(r0, r1, 1)})

    elif feature.__class__.__name__ in ['TrajectoryProfile']:
        i_axis = feature.ds.cb.cf_instance_dims[0]
        z_axis = feature.ds.cb.cf_axis_dims['Z']
        times = feature.ds.dims[i_axis]
        z = feature.ds.dims[z_axis]
        width = min(times // 2, 5)
        r0, r1 = times // 2 - width, times // 2 + width
        subset = feature.isel(
            {i_axis: slice(r0, r1, 1), z_axis: slice(0, z // 2)})

    elif feature.__class__.__name__ in ['TimeSeriesProfile']:
        i_axis = feature.ds.cb.cf_instance_dims[0]
        z_axis = feature.ds.cb.cf_axis_dims['Z']
        times = feature.ds.dims[i_axis]
        z = feature.ds.dims[z_axis]
        width = min(times // 2, 5)
        r0, r1 = times // 2 - width, times // 2 + width
        subset = feature.isel(
            {i_axis: slice(r0, r1, 1), z_axis: slice(0, z // 2)})

    elif feature.__class__.__name__ in ['IMDCollection', 'OMDCollection']:
        instance_dim  = feature.ds.cb.cf_instance_dims[0]
        instances = feature.ds.dims[instance_dim]
        width = min(instances // 2, 5)
        r0, r1 = instances // 2 - width, instances // 2 + width
        if r0 == r1:
            r1 += 1
        subset = feature.isel(
            {instance_dim: slice(r0, r1, 1)})

    else:
        raise NotImplementedError

    assert isinstance(subset, feature_class)

    return subset


def test_extract_and_save_a_subset(feature, feature_class, output_dir):
    """Test extracting a subset from a file with feature and saving it"""
    # extract
    subset = __extract_subset(feature, feature_class)

    # test saving the subset in netCDF format
    fname = output_dir / TEST_SUBSET
    print(f'Output dir: {fname}')

    subset.dataset.cb.save(fname)

    # try reading the subset to check it si correctly formatted
    featureobj = feature_class(xr.open_dataset(fname))
    assert isinstance(featureobj, feature_class)



# -*- coding: utf-8 -*-
"""
Class for the swath feature
"""
from __future__ import absolute_import, division, print_function

from cerbere.feature.feature import Feature

__all__ = ['Swath']


class Swath(Feature):
    """
    Feature class for representing a swath, a two-dimensional irregular grid
    along the satellite track.
    """
    _feature_geodimnames = 'row', 'cell',

    def __init__(self, *args, **kwargs):
        # create feature
        super(Swath, self).__init__(
            *args,
            **kwargs
        )

    def get_geocoord_dimnames(self, fieldname, shape=None):
        if fieldname == 'depth':
            return 'depth',
        else:
            return 'row', 'cell',



#     def extract_subset(
#             self, boundaries=None, slices=None, fields=None, padding=False,
#             prefix=None):
#         """Extract a subset feature from the swath.
#
#         The created subset is a new Swath object without any reference to
#         the source.
#
#         Args:
#             boundaries (tuple): area of the subset to extract, defined as
#                 llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat.
#
#             slices (dict): indices for /time/ dimension of the subset to
#                 extract from the source data. If None, the complete feature
#                 is extracted.
#
#             fields (list): list of field names to extract. If None, all fields
#                 are extracted.
#
#             padding (bool): Passed to extract_field method to ensure padding
#                 with _FillValues for points outside of the bounds of the this
#                 feature (used only in conjuncture with slices.
#
#             prefix (str): add a prefix string to the field names of the
#                 extracted subset.
#         """
#         if boundaries and slices:
#             raise ValueError("Boundaries and slices can not be both provided.")
#         if boundaries:
#             # get corresponding slices
#             raise ValueError("Not yet implemented")
#         elif slices:
#             if len(slices) != 2:
#                 raise ValueError(
#                     "slices for the two spatial dimensions of "
#                     "the swath must be provided"
#                 )
#
#         # fields to extract
#         extr_fields = OrderedDict([])
#         for f in GEOCOORDINATES:
#             if f in self.get_geolocation_fields():
#                 extr_fields[f] = self.extract_field(
#                     f, slices=slices, padding=padding
#                 )
#         sub_swath = Swath(
#             latitudes=self.extract_field(
#                 'lat', slices=slices, padding=padding),
#             longitudes=self.extract_field(
#                 'lon', slices=slices, padding=padding
#             ),
#             times=self.extract_field('time', slices=slices, padding=padding),
#             depths=self.extract_field('z', slices=slices, padding=padding),
#             attrs=self.xrdataset.attrs
#         )
#
#         if fields is None:
#             fields = self.get_fieldnames()
#         elif not isinstance(fields, list):
#             raise TypeError("fields must be a list")
#         for field in fields:
#             if field not in self.get_fieldnames():
#                 raise ValueError("Field %s not existing" % field)
#             sub_swath.add_field(
#                 self.extract_field(field,
#                                    slices=slices,
#                                    padding=padding,
#                                    prefix=prefix)
#             )
#         return sub_swath

    # def get_spatial_resolution(self):
    #     """Return the spatial resolution of the feature"""
    #     if self.spatial_resolution is None:
    #         i = self.get_geolocation_dimsizes()['cell'] / 2
    #         j = self.get_geolocation_dimsizes()['row'] / 2
    #         res = Point.get_distance(
    #             Point(self.get_lon(
    #                 slices={'row': slice(j, j + 1, 1),
    #                         'cell': slice(i, i + 1, 1)}
    #             ),
    #                 self.get_lat(
    #                 slices={'row': slice(j, j + 1, 1),
    #                         'cell': slice(i, i + 1, 1)}
    #             )),
    #             Point(self.get_lon(
    #                 {'row': slice(j + 1, j + 2, 1),
    #                  'cell': slice(i + 1, i + 2, 1)}
    #             ),
    #                 self.get_lat(
    #                 {'row': slice(j + 1, j + 2, 1),
    #                  'cell': slice(i + 1, i + 2, 1)}
    #             ))
    #         ).ravel()[0] / numpy.sqrt(2)
    #         return res



"""
Base classes for cerbere dataset objects and to read/save datasets in different
file formats and conventions.

.. autosummary::
    :toctree: _autosummary

    field
    dataset
    ncdataset
    ghrsstncdataset
"""
import logging
from enum import Enum
from typing import Union

import numpy as np
import xarray as xr

# cerbere internals for encoding/decoding data to file or memory
ENCODING = 'cerbere'


class Encoding(Enum):
    """attributes for saving the encoding of a source file"""

    # source file's saved encoding attributes
    IO_FILLVALUE: str = 'io__FillValue'
    IO_SCALE: str = 'io_scale_factor'
    IO_OFFSET: str = 'io_add_offset'
    IO_DTYPE: str = 'io_dtype'

    # in memory numpy encoding
    M_DTYPE: str = 'm_dype'
    M_MASK: Union[np.ndarray, bool] = 'mask'

    # individual components for vectorial fields
    COMPONENTS: str = 'components'

    # field status wrt to its copy on file
    STATUS: str = 'status'


def default_fill_value(obj):
    """Returns the default fill value for a specific type"""
    if isinstance(obj, np.dtype):
        dtype = obj
    elif isinstance(obj, (str, type)):
        dtype = np.dtype(obj)
    elif isinstance(obj, np.ndarray):
        dtype = obj.dtype
    else:
        raise TypeError('Unexpected object type: ', type(obj), obj)

    if dtype.name == 'int16':
        return np.int16(-32768)
    elif dtype.name == 'uint16':
        return np.uint16(65535)
    elif dtype.name == 'int8':
        return np.int8(-128)
    elif dtype.name == 'uint8':
        return np.uint8(255)
    elif dtype.name == 'float32':
        return np.float32(np.ma.default_fill_value(dtype))
    elif dtype.name == 'float64':
        return np.float64(np.ma.default_fill_value(dtype))
    else:
        return np.ma.default_fill_value(dtype)


def infer_fillvalue(fill_value, data, dtype):
    if fill_value is not None:
        return fill_value

    # use valid xarray missing values if possible (float or datetime64)
    if np.issubdtype(data.dtype, np.floating) or \
            np.issubdtype(data.dtype, np.complexfloating):
        fill_value = np.nan
    elif np.issubdtype(dtype, np.datetime64):
        fill_value = np.datetime64('NaT')
    else:
        try:
            fill_value = data.fill_value
        except:
            fill_value = default_fill_value(dtype)

    return fill_value

def to_cerbere_dataarray(
        data: Union[xr.DataArray, np.ndarray, np.ma.MaskedArray],
        fill_value: float = None,
        dtype: np.dtype = None,
        no_missing_value: bool = False,
        **kwargs) -> xr.DataArray:
    """Converts to a cerbere DataArray.

    A cerbere DataArray adds specific internal attributes into the
    `encoding` property of a DataArray, such as:
      * the mask of missing values (like for a numpy MaskedArray),
        which is not supported by DataArray (which replaces masked values
        with NaN)
      * the scientific dtype of the data (which may be different for the in
        memory encoding since xarray will transform the dtype of MaskedArray
        into floats.

    Used for internal data encoding as xarray data can't store masked
    values or nan for non-float types and transforms masked arrays to float
    arrays, which leads to a changed dtype of an array. When using cerbere
    `get_values` method, the data returned to the users will be in the
    internal DataArray dtype (with NaN for fill values when any), unless
    `decoding` keyword is set to True (NaN are then replaced be the fill_value
    attribute and the dtype of the returned values is the scientific one). The
    mask can be retrieved with the `mask` property of a Field object.

    Args:
        data: Array to be converted to masked DataArray
        fill_value: Fill value. Must be of the same type as dtype
        dtype: array's scientific data type. It can be different
            from `dtype` of `data` array. Data returned with `get_values`
            will be forced to this type.
        **kwargs: other DataArray creation arguments
    """
    if dtype is None:
        dtype = data.dtype

    if dtype != data.dtype:
        logging.debug('Data internally stored as {} but tht scientific dtype '
                      'is {}'.format(data.dtype, dtype))

    if no_missing_value and fill_value is not None:
        raise ValueError('no_missing_value is set to True but a fill_value was '
                         'also provided')

    if fill_value is None and data.dtype != dtype:
        # There is no reason here for xarray to have a different data type
        # than the scientific one.
        logging.warning('internal data type {} should be the same than '
                        'scientific data type {}'.format(data.dtype, dtype))

    if isinstance(data, np.ma.core.MaskedArray):
        if no_missing_value:
            if data.count() != data.size:
                raise ValueError('no_missing_value set to True but data '
                                 'contains masked values')
            logging.warning('data was provided as a masked array with '
                            'no_missing_value set to True. data will be '
                            'forced to non masked array.')
            mask = None
        else:
            mask = data.mask
    else:
        # mask - case ndarray, Dask array DataArray
        try:
            # numpy array type
            mask = np.isnan(data) | (data == fill_value)
        except:
            # xarray array type
            mask = data.isnull() | (data == fill_value)
        if np.all(mask):
            mask = True
        elif np.all(~mask):
            mask = False
            if fill_value is None:
                # no explicit fill values, no nan in data : this is a unmasked
                # variable
                mask = None

    # wrap into xarray DataArray
    data = xr.DataArray(data=data, **kwargs)

    if ENCODING not in data.encoding:
        data.encoding[ENCODING] = {}
    data.encoding[ENCODING][Encoding.M_DTYPE] = dtype

    # no mask in the variable
    data.encoding[ENCODING][Encoding.M_MASK] = mask

    # ensure there is a defined fill value
    if mask is not None:
        data.encoding['_FillValue'] = infer_fillvalue(
            fill_value, data, dtype)

    return data


def io_encoding_dtype(data: xr.DataArray) -> np.dtype:
    """guess the scientific dtype from data read on file"""
    scale = data.encoding.get(
        'scale_factor', data.encoding.get('add_offset', None))
    fillv = data.encoding.get(
        '_FillValue', data.attrs.get('_FillValue', None))
    if scale is None:
        # no scaling of data -> dtype should be unchanged

        if fillv is None:
            return data.dtype

        if np.dtype(type(fillv)) != data.dtype and not np.issubdtype(
                data.dtype, np.datetime64):
            # no scaling was applied but xarray may have changed the dtype for
            # instance if fill values were replaced with NaN. Returns the
            # intended scientific dtype indicated by the _FillValue attr.
            return np.dtype(type(fillv))

        return data.dtype

    else:
        dtype = np.dtype(type(scale))
        if dtype != data.dtype:
            # xarray may have changed the dtype for instance if fill values
            # were replaced with NaN. Returns the intended scientific dtype
            # indicated by the scaling attributes
            return dtype

    return dtype


def from_cerbere_dataarray(
        data,
        as_masked_array: bool = True,
        decoding: bool = False) -> Union[xr.DataArray, np.ma.MaskedArray]:
    """"""
    if not isinstance(data, xr.DataArray):
        raise TypeError('unexpected data type: {}'.format(type(data)))

    fillv = data.encoding.get('_FillValue', data.attrs.get(
        '_FillValue', None))

    if ENCODING not in data.encoding or \
            Encoding.M_DTYPE not in data.encoding[ENCODING]:
        # the data were not yet internally encoded or were read from an
        # externally defined DataArray or from a file => us the native encoding
        if as_masked_array:
            dtype = io_encoding_dtype(data)
            data = data.astype(dtype, copy=False).to_masked_array(copy=False)
            data.set_fill_value(fillv)
        elif decoding:
            dtype = io_encoding_dtype(data)
            data = data.astype(dtype, copy=False)

        return data

    if decoding and not as_masked_array:
        # return the data in their scientific dtype, replacing NaN with fill
        # values
        return data.fillna(data.encoding['_FillValue']).astype(
            data.encoding[ENCODING][Encoding.M_DTYPE], copy=False)

    if as_masked_array:
        mdata = data.astype(
            data.encoding[ENCODING][Encoding.M_DTYPE],
            copy=False).to_masked_array(copy=False)

        if Encoding.M_MASK in data.encoding[ENCODING]:
            mask = data.encoding[ENCODING][Encoding.M_MASK]
            if mask is not None:
                mdata = np.ma.masked_where(mask, data)
        else:
            mdata.mask = np.where(mdata == fillv)

        mdata.set_fill_value(fillv)

        return mdata

    # return the data in their internal dtype
    return data



"""
cerbere classes for common observation patterns.

.. autosummary::
    :toctree: _autosummary

    feature
    swath
    grid
    trajectory
"""
from cerbere.feature.cpoint import Point
from cerbere.feature.cprofile import Profile
from cerbere.feature.ctimeseries import TimeSeries
from cerbere.feature.ctrajectory import Trajectory
from cerbere.feature.ctimeseriesprofile import (TimeSeriesProfile,
                                                UniZTimeSeriesProfile)
from cerbere.feature.ctrajectoryprofile import TrajectoryProfile
from cerbere.feature.cimdcollection import IMDCollection
from cerbere.feature.comdcollection import OMDCollection
from cerbere.feature.cimage import Image
from cerbere.feature.cswath import Swath
from cerbere.feature.cgrid import Grid
from cerbere.feature.cgridtimeseries import GridTimeSeries



# -*- coding: utf-8 -*-
"""
Definition of Cerbere Climate & Forecast convention profile
"""
import collections
from enum import Enum
import inspect
import logging
import os
from typing import Any, Mapping

import yaml

import numpy as np


CF_AUTHORITY = 'CF-1.11, ACDD-1.3'

# Module level default values
DEFAULT_TIME_UNITS = 'seconds since 1981-01-01 00:00:00'

# format for printing times
FORMAT_TIME = '%Y-%m-%dT%H:%M:%SZ'

# standard CF structuring dimensions
CF_DIMS = [u'time', u'lat', u'lon', u'z', u'obs', u'profile', u'row', u'cell']

# Standard CF Axis attribute values
CF_AXIS = ['X', 'Y', 'Z', 'T']

# Standard CF coordinates
CF_COORDS = ['latitude', 'longitude', 'vertical', 'time']

# Cerbere naming for CF coordinate
CB_COORD_CONVENTION = {'latitude': 'lat', 'longitude': 'lon', 'time': 'time'}

CF_COORD_AXIS = {
    'latitude': 'Y',
    'longitude': 'X',
    'time': 'T',
    'vertical': 'Z'}

# Standard CF coordinate units
CF_COORD_UNITS = {
    'lat': ['degrees_north', 'degree_north', 'degree_N', 'degrees_N',
            'degreeN', 'degreesN'],
    'lon': ['degrees_east', 'degree_east', 'degree_E', 'degrees_E', 'degreeE',
            'degreesE'],
}

# standard geolocation coordinates
CF_REQ_GEOCOORDS = ['time', 'latitude', 'longitude']
GEOCOORDS = CF_REQ_GEOCOORDS + ['vertical']

GEOCOORD_AXIS = {
    'latitude': 'Y', 'longitude': 'X', 'time': 'T', 'vertical': 'Z'}

# standard attributes for geolocation coordinates
GEOCOORD_ATTRS = {
    'latitude': {
        'standard_name': 'latitude', 'units': 'degree_north',
        'long_name': 'latitude', 'authority': CF_AUTHORITY},
    'longitude': {
        'standard_name': 'longitude', 'units': 'degree_east',
        'long_name': 'longitude', 'authority': CF_AUTHORITY},
    'time': {
        'standard_name': 'time', 'long_name': 'time',
        'authority': CF_AUTHORITY},
}

# CF standard names for geolocation coordinates
GEOCOORD_STD_NAME = {
    'latitude': ['latitude'],
    'longitude': ['longitude'],
    'time': ['time'],
    # remove any value in 'vertical' as it is too ambiguous with non
    # coodinate variables
    'vertical': [],
}

# coverage_content_type (ACDD 1.3): an ISO 19115-1 code to indicate the
# source of the data.
COVERAGE_CONTENT_TYPE = [
    'image', 'thematicClassification', 'physicalMeasurement',
    'auxiliaryInformation', 'qualityInformation', 'referenceInformation',
    'modelResult', 'coordinate'
]

CDM2CF_FEATURE = {
    'grid': None,
    'swath': None,
    'curvilinear': None,
    'point': 'point',
    'profile': 'profile',
    'station': 'timeSeries',
    'trajectory': 'trajectory',
    'station_profile': 'timeSeriesProfile',
    'trajectory_profile': 'trajectoryProfile'
}

"""
public enum ucar.nc2.constants.FeatureType {
  ANY,        // No specific type

  COVERAGE,   // any of the coverage types: GRID, FMRC, SWATH, CURVILINEAR
  GRID,       // seperable coordinates
  FMRC,       // two time dimensions, runtime and forecast time
  SWATH,      // 2D latlon, dependent time, polar orbiting satellites
  CURVILINEAR,// 2D latlon, independent time

  ANY_POINT,  // Any of the point types
  POINT,      // unconnected points
  PROFILE,    // fixed x,y with data along z
  STATION,    // timeseries at named location
  STATION_PROFILE, // timeseries of profiles
  TRAJECTORY, // connected points in space and time
  TRAJECTORY_PROFILE, //  trajectory of profiles

  RADIAL,     // polar coordinates
  STATION_RADIAL, // time series of radial data

  // experimental
  IMAGE,    // pixels, may not be geolocatable
  UGRID;    // unstructured grids
}
"""


def default_fill_value(obj):
    """Returns the default fill value for a specific type"""
    if isinstance(obj, np.dtype):
        dtype = obj
    elif isinstance(obj, (str, type)):
        dtype = np.dtype(obj)
    elif isinstance(obj, np.ndarray):
        dtype = obj.dtype
    else:
        raise TypeError('Unexpected object type: ', type(obj), obj)

    if np.issubdtype(dtype, np.datetime64):
        return np.datetime64('NaT')

    if np.issubdtype(dtype, np.integer):
        # return the maximum for numbers, the minimum for relative numbers
        info = np.iinfo(dtype)
        if info.min == 0:
            return dtype.type(info.max)
        return dtype.type(info.min)

    return dtype.type(np.ma.default_fill_value(dtype))


def default_profile(
        profile: str='default_saving_profile.yaml') -> Mapping[str, Any]:
    """Returns a list of default settings for storing data and metadata

    The settings are defined in a YAML file and can include a liste of global
    attributes (and default values), variable attributes (and default values)
    and variable encoding (scale_factor, add_offset,...) for a given storage
    format. If no profile file is provided, the function returns the built-in
    default settings provided in the ``cerbere.share`` folder of the package.

    These settings can be customized by defining your own in a new YAML profile
    file and calling the function with ``profile`` argument. If the profile file
    is stored in the user home dir in ``~.cerbere`` folder, providing only the
    file basename is fine, otherwise provide the full path to the profile file.

    A custom file can be defined in YAML, using the following format:

    .. code-block:: yaml

        ---
        # Defines first the global attributes
        attributes:
            # define here a dictionary of global attributes (possibly with
            # default value in the following form
            #
            # with default value
            attr1: val1

            # with no default value (None)
            attr2:

        # Define the variable attributes
        variables:
            # a dictionary of variables containing a dictionary of variable
            # attributes
            variable1:
                varattr1: varval1
                varattr2:

        # Define the encoding attributes, meaningful for a specific format. For
        # instance for NetCDF4, allowed encoding attributes include:
        # scale_factor, add_offset, dtype, least_significant_digit,...
        encoding:
            # a dictionary of variables containing a dictionary of encoding
            # attributes
            variable1:
                scale_factor: 0.01
                add_offset: 273.15
                dtype: int32

    Args:
        profile: the path to (or filename of) the settings profile file

    Returns:
        a dict of attributes and values
    """
    if profile is None:
        profile = 'default_saving_profile.yaml'
    if not os.path.exists(profile):
        # get path from home dir or default
        path = os.path.join(
            os.path.expanduser('~'),
            '.cerbere',
            profile
        )
        if not os.path.exists(path):
            path = os.path.join(
                os.path.dirname(inspect.getfile(inspect.currentframe())),
                'share/{}'.format(profile)
            )
            logging.warning(
                'Using default global attribute file: {}'.format(path)
            )
    else:
        path = profile

    # read attributes
    with open(os.path.abspath(path), 'r') as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
            raise

    globattrs = collections.OrderedDict([])
    varattrs = collections.OrderedDict([])
    encoding = collections.OrderedDict([])

    if 'attributes' in config:
        for att in config['attributes']:
            globattrs[att] = config['attributes'][att]
    if 'fields' in config and config['fields'] is not None:
        for v in config['fields']:
            varattrs[v] = config['fields'][v]
    if 'encoding' in config and config['encoding'] is not None:
        for v in config['encoding']:
            encoding[v] = config['encoding'][v]

    return globattrs, varattrs, encoding



# -*- coding: utf-8 -*-
"""
Feature classes for grid observation patterns
"""
import copy
import inspect
import typing as T
import warnings

import xarray as xr
from xarray.core.utils import either_dict_or_kwargs

import cerbere
from cerbere.feature.cbasefeature import BaseFeature


__all__ = ['Grid']


# Plate Carree projection as default (regular isolat, isolon)
DEFAULT_PROJECTION = xr.DataArray(
    data=0,
    attrs={
        'grid_mapping_name': "latitude_longitude",
        'longitude_of_prime_meridian': 0.0,
        'semi_major_axis': 6378137.0,
        'inverse_flattening': 298.257223563}
)


class Grid(BaseFeature):
    """
    Model class for the grid feature, ie a two-dimensional array on fixed
    projection, resolution and boundaries.

    A grid can be built from the ``content`` argument:
        * if None, the feature will undefined but can be updated later (for
          instance loading the content from a file with ``load`` method)
        * from the content of a file, providing a mapper class object
        * from an xarray dataset object
        * from a dictionary of coordinate and data fields. For coordinate
          fields, use only any of 'lat', 'lon', 'time', 'z' as keys.

          The coordinate fields can be provided as numpy arrays of values,
          :class:`cerbere.datamodel.field.Field` class objects, or
          :class:`xarray.DataArray` class objects.

          The lat and lon coordinate fields can also be provided as tuple
          (min, max, step): this is only for a regular grid. The values
          are automatically generated.

          The geophysical fields, in ``fields`` can be
          :class:`cerbere.datamodel.field.Field` class objects, or
          :class:`xarray.DataArray` class objects.

    .. code-block:: python

        from cerbere.feature.grid import Grid
        from cerbere.dataset.ncdataset import NCDataset

        ncf = NCDataset(url='SSS_SMOS_L3_Daily_0.5deg_CATDS_CECOS_2012.12.30_V02.nc')
        g = Grid(ncf)

    Args:
        content (xarray, mapper or dict of fields): the content from which to
            build the feature coordinates and data fields.

        attrs (dict): a dictionary of global attributes (as in a NetCDF file).
            Make use of the CF attributes. If the feature content is read from
            a file through a mapper, these attributes will be added to the
            existing the list of attributes already in the file (or replace the
            values of the existing attributes with the same name in the file).
            Use with caution in this context.
    """
    cdm_data_type = 'grid'

    # the expected shape of the coordinates and variables, as defined by CF
    # convention (CF 1.11, Table 9.1)
    cf_canonical_dims = {
        'longitude': [('X', 'Y'), ('X',)],
        'latitude': [('X', 'Y'), ('Y',)],
        'time': [('T',), ('X', 'Y',)],
        'vertical': ('Z',),
        'data': [('X', 'Y'), ('T', 'X', 'Y')]
    }
    # optional axes for the feature
    cf_optional_axes = ['vertical']

    def nearest(self, lon, lat, z=None) -> xr.Dataset:
        # @TODO
        pass

    def inearest(self, lon, lat, z=None) -> T.Tuple[int]:
        # @TODO
        pass

    @classmethod
    def check_feature(cls, dataset: xr.Dataset) -> bool:
        cfdst = dataset.cb.cfdataset

        if cfdst.cb.cdm_data_type is not None and \
                cfdst.cb.cdm_data_type.lower() != cls.cdm_data_type:
            return False

        # check grid mapping parameters
        if dataset.cb.grid_mapping is None:
            warnings.warn(
                'No grid mapping parameters in dataset',
                cerbere.CWarning)

        if cfdst.cb.grid_mapping is not None:
            return True

        # check lat/lon are axes and two-dimensional
        for coordname in ['lon', 'lat']:
            if len(cfdst[coordname].dims) == 1:
                return
        if set(cfdst['lon'].dims) != set(cfdst['lat'].dims):
            return

        # check provided axes if any
        if (cfdst.cb.X is not None
                and cfdst.cb.X.name == cfdst.cb.longitude.name):
            return
        if (cfdst.cb.Y is not None
                and cfdst.cb.Y.name == cfdst.cb.latitude.name):
            return

        # time is unidimensional or (X, Y)
        if (len(cfdst.time.dims) != 1 and
                set(cfdst.time.dims) != {cfdst.cb.X.name, cfdst.cb.Y.name}):
            return


    @classmethod
    def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[BaseFeature]:

        cfdst = dataset.cb.cfdataset

        # check CF featureType and Unidata CDM' feature type
        if (cfdst.cb.cdm_data_type is not None
                and not cls._unexpected_cdm_data_type(cfdst)):
            return Grid

        # check horizontal axes are defined or at least some mapping
        if cfdst.cb.X is None and cfdst.cb.Y is None:
            if cfdst.cb.grid_mapping is None or cfdst.cb.grid_mapping_name:
                return

        if (cfdst.cb.cf_axis_dims.get('X') is None
                or cfdst.cb.cf_axis_dims.get('Y') is None):
            return
        elif  cfdst.cb.cf_axis_dims['X'] == cfdst.cb.cf_axis_dims['Y']:
            return

        # time is unidimensional or (X, Y)
        if (len(cfdst.time.dims) != 1 and
                set(cfdst.time.dims) != {cfdst.cb.X.name, cfdst.cb.Y.name}):
            return

        return Grid

    @classmethod
    def cerberize(
            cls,
            dataset: xr.Dataset,
            force_reorder: bool = True,
            instance_class: 'BaseFeature' = None
    ) -> xr.Dataset:
        cfdst = super().cerberize(dataset, force_reorder, instance_class)

        # add missing default mapping for cylindrical grids
        if dataset.cb.grid_mapping is None and cls._is_cylindrical(dataset):
            cfdst['crs'] = DEFAULT_PROJECTION

        # add mapping information to data variables
        if cfdst.cb.grid_mapping is not None:
            for v in cfdst.data_vars:
                if cfdst.cb.cf_axis_dims['X'] in cfdst[v].dims \
                        and cfdst.cb.cf_axis_dims['Y'] in cfdst[v].dims \
                        and 'grid_mapping' not in cfdst[v].attrs:
                    cfdst[v].attrs['grid_mapping'] = cfdst.cb.grid_mapping.name

        # squeeze time dimension in non-coordinate fields
        ctime = cfdst.time.copy()
        if 'time' in cfdst.dims and cfdst.sizes['time'] <= 1:
            cfdst = cfdst.squeeze(dim='time')
        cfdst.coords['time'] = ctime

        # Unidata CDM feature type
        if cls._is_cylindrical(cfdst):
            cfdst.attrs['cdm_data_type'] = 'grid'
        else:
            cfdst.attrs['cdm_data_type'] = cls.cdm_data_type

        return cfdst

    @classmethod
    def _is_cylindrical(cls, dataset) -> bool:
        # check lat/lon are axes and unidimensional
        if dataset.cb.X is None or dataset.cb.Y is None:
            return False

        if (dataset.cb.X.name != dataset.cb.longitude.name or
                dataset.cb.Y.name != dataset.cb.latitude.name):
            return False

        return True

    def is_cylindrical(self) -> bool:
        return self._is_cylindrical(self.dataset)

    def is_circular(self, precision: int = 3) -> bool:
        """Return True if the grid is cylindrical and False otherwise."""
        if not self.is_cylindrical():
            return False

        return (
            ((self.ds.lon[0] - self.ds.lon[-1]) % 360.).round(precision) ==
            ((self.ds.lon[1] - self.ds.lon[0]) % 360.) .round(precision))

    def isel(
            self,
            indexers: T.Any = None,
            circularity: bool = True,
            **kwargs
    ) -> T.Any:
        """
        shortcut for cerbere ds.cb.isel method but returns the result
        wrapped into a Feature object of the same class as the current object

        Parameters
        ----------
        circularity : bool
            in case of negative slices in the longitude axis, combine
            westward and eastward slices apart from the start longitude axis.
            True by default and overrides padding if set.
        """
        # separate indexer kargs from isel other kwargs
        isel_signature = inspect.getfullargspec(self.ds.cb.isel).args
        isel_kwargs = {k: v for k, v in kwargs.items() if k in isel_signature}
        indx_kwargs = {k: v for k, v in kwargs.items() if k not in
                       isel_signature}
        indexers = either_dict_or_kwargs(
            copy.copy(indexers), indx_kwargs, "isel")

        if not self.is_cylindrical() and circularity:
            # can not be applied to non cylindrical grids
            circularity = False

        if (not circularity
                or not self.is_circular()
                or 'lon' not in indexers
                or not isinstance(indexers['lon'], slice)):
            return self.__class__(self.ds.cb.isel(**indexers, **isel_kwargs))

        lon_slice = indexers.pop('lon')
        if lon_slice.start >= 0 and lon_slice.stop < self.ds.lon.size:
            # no split necessary
            return self.__class__(self.ds.cb.isel(
                {**indexers, **{'lon': lon_slice}}, **isel_kwargs))

        if lon_slice.start < 0 and lon_slice.stop > self.ds.lon.size:
            raise ValueError(f'slice {lon_slice} can not self wrap over a grid')

        if lon_slice.start < 0:
            lon_slice = [
                slice(self.ds.lon.size + lon_slice.start, self.ds.lon.size),
                slice(0, lon_slice.stop)]
        else:
            lon_slice = [
                slice(lon_slice.start, self.ds.lon.size),
                slice(0, lon_slice.stop - self.ds.lon.size)]

        return self.__class__(xr.concat(
            [self.ds.cb.isel(
                **{**indexers, **{'lon': lon_slice[0]}}, **isel_kwargs),
             self.ds.cb.isel(
                 **{**indexers, **{'lon': lon_slice[1]}}, **isel_kwargs)],
            dim='lon'))





#
# class CylindricalGrid(Grid):
#     """
#     Cylindrical lat/lon grid.
#     """
#     cdm_data_type = 'grid'
#
#     # the expected shape of the coordinates and variables, as defined by CF
#     # convention (CF 1.11, Table 9.1)
#     cf_canonical_dims = {
#         'longitude': ('X',),
#         'latitude': ('Y',),
#         'time': [('T',), ('X', 'Y',)],
#         'vertical': ('Z',),
#         'data': ('X', 'Y',)
#     }
#     # optional axes for the feature
#     cf_optional_axes = ['vertical']
#
#     def nearest(self, lon, lat, z=None) -> xr.Dataset:
#         # @TODO
#         if z is None:
#             return self.dataset.cf.sel(lon=lon, lat=lat, method='nearest')
#
#         z_coord = self.dataset.cf.axis_coordname('Z')
#         return self.dataset.cf.sel(
#             **{'lon': lon, 'lat': lat, z_coord: z}, method='nearest')
#
#     def inearest(self, lon, lat, z=None) -> T.Tuple[int]:
#         # @TODO
#         pass
#
#     @classmethod
#     def guess_feature(cls, dataset: xr.Dataset) -> T.Optional[BaseFeature]:
#
#         cfdst = dataset.cb.cfdataset
#
#         # check Unidata CDM' feature type
#         if cfdst.cb.cdm_data_type is not None and \
#                 cfdst.cb.cdm_data_type.lower() != cls.cdm_data_type:
#             return
#
#         # check lat/lon are axes and unidimensional
#         for coordname in ['lon', 'lat']:
#             if len(cfdst[coordname].dims) != 1:
#                 return
#         if set(cfdst['lon'].dims) == set(cfdst['lat'].dims):
#             return
#
#         # check provided axes if any
#         if (cfdst.cb.X is not None
#                 and cfdst.cb.X.name != cfdst.cb.longitude.name):
#             return
#         if (cfdst.cb.Y is not None
#                 and cfdst.cb.Y.name != cfdst.cb.latitude.name):
#             return
#
#         # time is unidimensional or (lat, lon)
#         if (len(cfdst.time.dims) != 1 and
#                 set(cfdst.time.dims) != {cfdst.cb.X.name, cfdst.cb.Y.name}):
#             return
#
#         return CylindricalGrid
